@@ -17,7 +17,15 @@ class ReManagerEmulation(threading.Thread):
     """
 
     def __init__(
-        self, *args, conn_watchdog, conn_worker, config=None, msg_queue=None, log_level=logging.DEBUG, **kwargs
+        self,
+        *args,
+        conn_watchdog,
+        conn_worker,
+        config=None,
+        msg_queue=None,
+        log_level=logging.DEBUG,
+        number_of_restarts,
+        **kwargs,
     ):
         super().__init__(*args, **kwargs)
         self._conn_watchdog = conn_watchdog
@@ -28,6 +36,7 @@ class ReManagerEmulation(threading.Thread):
         self._lock = threading.Lock()
         self._config_dict = config or {}
         self._log_level = log_level
+        self._number_of_restarts = number_of_restarts
 
     def _heartbeat(self):
         hb_period, dt = 0.5, 0.01
@@ -100,7 +109,16 @@ class ReManagerEmulation(threading.Thread):
 
 
 class ReWorkerEmulation(threading.Thread):
-    def __init__(self, *args, conn, config=None, msg_queue=None, log_level=logging.DEBUG, **kwargs):
+    def __init__(
+        self,
+        *args,
+        conn,
+        config=None,
+        msg_queue=None,
+        log_level=logging.DEBUG,
+        user_group_permissions=None,
+        **kwargs,
+    ):
         super().__init__(*args, **kwargs)
         self._config_dict = config or {}
         self._exit = False
@@ -133,13 +151,17 @@ def test_WatchdogProcess_1():
 
 
 def test_WatchdogProcess_2():
-    """Test starting RE Manager, stopping heartbeat generator
-    and waiting for restart of RE Manager"""
+    """
+    Test starting RE Manager, stopping heartbeat generator and waiting for restart of RE Manager.
+    Check that the number of restarts of the manager process is properly counted and propagated.
+    """
     wp = WatchdogProcess(cls_run_engine_manager=ReManagerEmulation)
     wp_th = threading.Thread(target=wp.run)
     wp_th.start()
     ttime.sleep(1)  # Let RE Manager run 1 second
     assert wp._re_manager.n_loops > 0, "RE is not running"
+    assert wp._re_manager_n_restarts == 1
+
     n_loops = wp._re_manager.n_loops
 
     wp._re_manager.stop_heartbeat()
@@ -150,6 +172,7 @@ def test_WatchdogProcess_2():
     #   Here we check if RE Manager was really restarted and the number of
     #   loops reset.
     assert wp._re_manager.n_loops < n_loops, "Unexpected number of loops"
+    assert wp._re_manager_n_restarts == 2
 
     wp._re_manager.exit(restart=False)
     ttime.sleep(0.05)
@@ -193,18 +216,18 @@ def test_WatchdogProcess_4():
     wp_th.start()
     ttime.sleep(0.01)
 
-    response = wp._re_manager.send_msg_to_watchdog("start_re_worker")
-    assert response["success"] is True, "Unexpected response from RE Manager"
+    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
+    assert response["success"] is True, f"Unexpected response from RE Manager: {response}"
 
     # Worker is expected to be alive
     response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
-    assert response["worker_alive"] is True, "Unexpected response from RE Manager"
+    assert response["worker_alive"] is True, f"Unexpected response from RE Manager: {response}"
 
     # Join running process (thread). Expected to timeout.
     # Note: here timeout should be set to be smaller than timeout for the message
     #   in 'send_msg_to_watchdog method.
     response = wp._re_manager.send_msg_to_watchdog("join_re_worker", {"timeout": 0.1})
-    assert response["success"] is False, "Unexpected response from RE Manager"
+    assert response["success"] is False, f"Unexpected response from RE Manager: {response}"
 
     # Worker is expected to be alive
     response = wp._re_manager.send_msg_to_watchdog("is_worker_alive")
@@ -234,7 +257,7 @@ def test_WatchdogProcess_5():
     wp_th.start()
     ttime.sleep(0.01)
 
-    response = wp._re_manager.send_msg_to_watchdog("start_re_worker")
+    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
     assert response["success"] is True, "Unexpected response from RE Manager"
 
     # Worker is expected to be alive
@@ -273,7 +296,7 @@ def test_WatchdogProcess_6():
     wp_th.start()
     ttime.sleep(0.01)
 
-    response = wp._re_manager.send_msg_to_watchdog("start_re_worker")
+    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
     assert response["success"] is True, "Unexpected response from RE Manager"
 
     # Check if configuration was set correctly in RE Worker and RE manager
@@ -311,7 +334,7 @@ def test_WatchdogProcess_7():
     wp_th.start()
     ttime.sleep(0.01)
 
-    response = wp._re_manager.send_msg_to_watchdog("start_re_worker")
+    response = wp._re_manager.send_msg_to_watchdog("start_re_worker", params={"user_group_permissions": {}})
     assert response["success"] is True, "Unexpected response from RE Manager"
 
     # Check if configuration was set correctly in RE Worker and RE manager
